from typing import Any, Optional

from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext
from untypy.util.typehints import get_type_hints


class StringForwardRefFactory(TypeCheckerFactory):
    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if type(annotation) is str:
            eval_args = [annotation, globals()]
            local = ctx.eval_context()
            if local is not None:
                eval_args.append(local)

            def resolver(eval_args):
                return eval(*eval_args)

            annotation = get_type_hints(eval_args, ctx, resolver=resolver)
            return ctx.find_checker(annotation)



from typing import Any, Optional

from untypy.error import UntypyTypeError
from untypy.interfaces import TypeChecker, CreationContext, TypeCheckerFactory, ExecutionContext


class DummyDelayedType:
    """
    This class is used for raising delayed type checking errors.
    """
    pass


class DummyDelayedFactory(TypeCheckerFactory):
    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if annotation is DummyDelayedType:
            return DummyDelayedChecker()
        else:
            return None


class DummyDelayedChecker(TypeChecker):
    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        return DummyDelayedWrapper(ctx)

    def describe(self) -> str:
        return "DummyDelayedType"

    def base_type(self) -> list[Any]:
        return []


class DummyDelayedWrapper:
    upper: ExecutionContext

    def __init__(self, upper: ExecutionContext):
        self.upper = upper

    def use(self):
        raise self.upper.wrap(UntypyTypeError(
            "<omitted>",
            "DummyDelayedType"
        ))



from typing import Any, Optional, Union

from untypy.error import UntypyTypeError, UntypyAttributeError
from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext
from untypy.util import CompoundTypeExecutionContext

UnionType = type(Union[int, str])


class UnionFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if type(annotation) is UnionType:
            inner = []
            for arg in annotation.__args__:
                checker = ctx.find_checker(arg)
                if checker is None:
                    return None
                else:
                    inner.append(checker)

            return UnionChecker(inner, ctx)
        else:
            return None


class UnionChecker(TypeChecker):
    inner: list[TypeChecker]

    def __init__(self, inner: list[TypeChecker], ctx: CreationContext):
        # especially Protocols must be checked in a specific order.
        self.inner = sorted(inner, key=lambda t: -t.base_type_priority())
        dups = dict()
        for checker in inner:
            for base_type in checker.base_type():
                if base_type in dups:
                    raise ctx.wrap(UntypyAttributeError(f"{checker.describe()} is in conflict with "
                                                        f"{dups[base_type].describe()} "
                                                        f"in {self.describe()}. "
                                                        f"Types must be distinguishable inside one Union."
                                                        f"\nNote: Only one Protocol is allowed inside one Union. "
                                                        f"Classes could implement multiple Protocols by accident."
                                                        f"\nNote: Multiple Callables or Generics inside one Union are also unsupported."))
                else:
                    dups[base_type] = checker

    def check_and_wrap(self, arg: Any, upper: ExecutionContext) -> Any:
        idx = 0
        for checker in self.inner:
            ctx = UnionExecutionContext(upper, self.inner, idx)
            idx += 1
            try:
                return checker.check_and_wrap(arg, ctx)
            except UntypyTypeError as _e:
                pass

        raise upper.wrap(UntypyTypeError(
            arg,
            self.describe()
        ))

    def describe(self) -> str:
        desc = lambda s: s.describe()
        return f"Union[{', '.join(map(desc, self.inner))}]"

    def base_type(self) -> list[Any]:
        out = []
        for checker in self.inner:
            out.append(checker.base_type())
        return out


class UnionExecutionContext(CompoundTypeExecutionContext):
    def name(self):
        return "Union"



import sys
import doctest
import os
import argparse
from dataclasses import dataclass
from runner import runCode, importUntypy, verbose, enableVerbose

usage = """python3 replTester.py [ ARGUMENTS ] LIB_1 ... LIB_n --repl SAMPLE_1 ... SAMPLE_m

If no library files should be used to test the REPL samples, omit LIB_1 ... LIB_n
and the --repl flag.
The definitions of LIB_1 ... LIB_n are made available when testing
SAMPLE_1 ... SAMPLE_m, where identifer in LIB_i takes precedence over identifier in
LIB_j if i > j.
"""

@dataclass
class Options:
    verbose: bool
    diffOutput: bool
    libs: list[str]
    repls: list[str]

def parseCmdlineArgs():
    parser = argparse.ArgumentParser(usage=usage,
                        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--verbose', dest='verbose', action='store_const',
                        const=True, default=False,
                        help='Be verbose')
    parser.add_argument('--diffOutput', dest='diffOutput',
                        action='store_const', const=True, default=False,
                        help='print diff of expected/given output')
    args, restArgs = parser.parse_known_args()
    libs = []
    repls = []
    replFlag = '--repl'
    if replFlag in restArgs:
        cur = libs
        for x in restArgs:
            if x == replFlag:
                cur = repls
            else:
                cur.append(x)
    else:
        repls = restArgs
    if len(repls) == 0:
        print('No SAMPLE arguments given')
        sys.exit(1)
    return Options(verbose=args.verbose, diffOutput=args.diffOutput, libs=libs, repls=repls)

opts = parseCmdlineArgs()

if opts.verbose:
    enableVerbose()

libDir = os.path.dirname(__file__)
libFile = os.path.join(libDir, 'writeYourProgram.py')
defs = globals()
importUntypy()
# runCode(libFile, defs, [])

for lib in opts.libs:
    d = os.path.dirname(lib)
    if d not in sys.path:
        sys.path.insert(0, d)

for lib in opts.libs:
    verbose(f"Loading lib {lib}")
    runCode(lib, defs, [])

totalFailures = 0
totalTests = 0

doctestOptions = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS

if opts.diffOutput:
    doctestOptions = doctestOptions | doctest.REPORT_NDIFF

# We use our own DocTestParser to replace exception names in stacktraces
class MyDocTestParser(doctest.DocTestParser):
    def get_examples(self, string, name='<string>'):
        prefs = {'WyppTypeError: ': 'untypy.error.UntypyTypeError: ',
                 'WyppNameError: ': 'untypy.error.UntypyNameError: ',
                 'WyppAttributeError: ': 'untypy.error.UntypyAttributeError: '}
        lines = []
        for l in string.split('\n'):
            for pref,repl in prefs.items():
                if l.startswith(pref):
                    l = repl + l[len(pref):]
            lines.append(l)
        string = '\n'.join(lines)
        x = super().get_examples(string, name)
        return x

for repl in opts.repls:
    (failures, tests) = doctest.testfile(repl, globs=defs, module_relative=False,
                                         optionflags=doctestOptions, parser=MyDocTestParser())

    totalFailures += failures
    totalTests += tests
    if failures == 0:
        if tests == 0:
            print(f'No tests in {repl}')
        else:
            print(f'All {tests} tests in {repl} succeeded')
    else:
        print(f'ERROR: {failures} out of {tests} in {repl} failed')

if totalFailures == 0:
    if totalTests == 0:
        print('ERROR: No tests found at all!')
        sys.exit(1)
    else:
        print(f'All {totalTests} tests succeded. Great!')
else:
    print(f'ERROR: {failures} out of {tests} failed')
    sys.exit(1)



import sys
from typing import Any, Callable

from untypy.error import Location, UntypyAttributeError, UntypyTypeError, Frame, UntypyNameError
from untypy.impl import DefaultCreationContext
from untypy.interfaces import ExecutionContext


class StandaloneChecker:
    def __init__(self, annotation: Callable[[], Any], declared: Any, cfg):
        self._checker = None
        self.annotation = annotation
        self.declared = declared
        self.cfg = cfg

    def get_checker(self):
        if self._checker:
            return self._checker

        ctx = DefaultCreationContext(
            typevars=dict(),
            declared_location=Location.from_code(self.declared),
            checkedpkgprefixes=self.cfg.checkedprefixes)
        try:
            annotation = self.annotation()
        except NameError as ne:
            raise ctx.wrap(UntypyNameError(
                f"{ne}.\nType annotation could not be resolved."
            ))

        checker = ctx.find_checker(annotation)
        if checker is None:
            raise ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {self.annotation}\n"))
        self._checker = checker
        return checker

    def __call__(self, val):
        frame = sys._getframe(2)
        checker = self.get_checker()
        ctx = StandaloneCheckerContext(frame, self.declared)
        return checker.check_and_wrap(val, ctx)

    def __repr__(self):
        return f"<StandaloneChecker for {self.get_checker().describe()}>"


class StandaloneCheckerContext(ExecutionContext):
    def __init__(self, caller, declared):
        self.caller = caller
        self.declared = declared

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        t, i = err.next_type_and_indicator()
        return err.with_frame(Frame(
            t, i,
            responsable=Location.from_stack(self.caller),
            declared=Location.from_code(self.declared)
        ))


import inspect
import sys
from types import GenericAlias
from typing import Any, Optional, List

from untypy.error import UntypyTypeError, Frame, Location
from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext


class ListFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if (type(annotation) is GenericAlias and annotation.__origin__ == list) or (
                type(annotation) is type(List[int]) and annotation.__origin__ == list):
            assert len(annotation.__args__) == 1
            inner = ctx.find_checker(annotation.__args__[0])
            if inner is None:
                return None
            return ListChecker(inner, ctx.declared_location())
        else:
            return None


class ListChecker(TypeChecker):
    inner: TypeChecker
    declared: Location

    def __init__(self, inner: TypeChecker, declared: Location):
        self.inner = inner
        self.declared = declared

    def may_change_identity(self) -> bool:
        return True

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if not issubclass(type(arg), list):
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

        return TypedList(arg, self.inner, ListExecutionContext(ctx), self.declared)

    def base_type(self) -> list[Any]:
        return [list]

    def describe(self) -> str:
        return f"list[{self.inner.describe()}]"


class ListExecutionContext(ExecutionContext):
    upper: ExecutionContext

    def __init__(self, upper: ExecutionContext):
        self.upper = upper

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        next_type, indicator = err.next_type_and_indicator()

        err = err.with_frame(Frame(
            f"list[{next_type}]",
            (" " * len("list[") + indicator),
            None,
            None
        ))
        return self.upper.wrap(err)


class ListCallerExecutionContext(ExecutionContext):
    stack: inspect.FrameInfo
    declared: Location

    def __init__(self, stack: inspect.FrameInfo, declared: Location):
        self.stack = stack
        self.declared = declared

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        next_type, indicator = err.next_type_and_indicator()
        return err.with_frame(Frame(
            f"list[{next_type}]",
            (" " * len("list[") + indicator),
            declared=self.declared,
            responsable=Location.from_stack(self.stack)
        ))

# ATTENTION: this is danger zone here. We had many bugs and these bugs are very ugly.
# Most methods are implemented in the same way as in collections.UserList. Hence, the ordering
# of methods is as in collections.UserList.
class TypedList(list):
    inner: list
    checker: TypeChecker
    ctx: ExecutionContext
    declared: Location

    def __init__(self, lst, checker, ctx, declared):
        super().__init__()
        self.checker = checker
        self.inner = lst
        self.ctx = ctx
        self.declared = declared

    def __repr__(self):
        return repr(self.inner)

    def __str__(self):
        return str(self.inner)

    # <
    def __lt__(self, other):
        return self.inner < other

    # <=
    def __le__(self, other):
        return self.inner <= other

    # ==
    def __eq__(self, other):
        return self.inner == other

    # !=
    def __ne__(self, other):
        return self.inner != other

    # >
    def __gt__(self, other):
        return self.inner > other

    # >=
    def __ge__(self, other):
        return self.inner >= other

    def __contains__(self, item):
        return self.inner.__contains__(item)

    def __len__(self):
        return len(self.inner)

    # Perform type check
    def __getitem__(self, index):
        if type(index) is int:
            # list[1], flat get
            ret = self.inner.__getitem__(index)
            return self.checker.check_and_wrap(ret, self.ctx)
        else:
            # returned structure is an list itself.
            # e.g. list[1:3, ...]
            return list(map(lambda x: self.checker.check_and_wrap(x, self.ctx), self.inner.__getitem__(index)))

    # l[i] = x
    # l[i:j] = [...]
    def __setitem__(self, idx, value):
        caller = sys._getframe(1)
        ctx = ListCallerExecutionContext(caller, self.declared)
        if isinstance(idx, slice):
            self.inner.__setitem__(idx, list(map(lambda x: self.checker.check_and_wrap(x, ctx), value)))
        else:
            return self.inner.__setitem__(idx, self.checker.check_and_wrap(value, ctx))

    def __delitem__(self, i):
        del self.inner[i]

    # l + ...
    def __add__(self, other):
        # Caller Context
        return self.inner + other

    # ... + l
    def __radd__(self, other):
        return other + self.inner

    # l += [...]
    def __iadd__(self, other):
        caller = sys._getframe(1)
        ctx = ListCallerExecutionContext(caller, self.declared)
        self.inner.extend(list(map(lambda x: self.checker.check_and_wrap(x, ctx), other)))
        return self

    def __mul__(self, n):
        return self.inner * n

    __rmul__ = __mul__

    def __imul__(self, n):
        return self.inner * n

    def __copy__(self):
        return self.inner.copy()

    def append(self, x) -> None:
        caller = sys._getframe(1)
        ctx = ListCallerExecutionContext(caller, self.declared)
        self.inner.append(self.checker.check_and_wrap(x, ctx))

    def insert(self, index, obj) -> None:
        caller = sys._getframe(1)
        ctx = ListCallerExecutionContext(caller, self.declared)
        return self.inner.insert(index, self.checker.check_and_wrap(obj, ctx))

    def pop(self, i=-1):
        return self.inner.pop(i)

    def remove(self, item):
        return self.inner.remove(item)

    def clear(self):
        return self.inner.clear()

    def copy(self):
        return self.inner.copy()

    def count(self, item):
        return self.inner.count(item)

    def index(self, *args):
        return self.inner.index(*args)

    def reverse(self):
        return self.inner.reverse()

    def sort(self, /, *args, **kwargs):
        return self.inner.sort(*args, **kwargs)

    def extend(self, iterable) -> None:
        caller = sys._getframe(1)
        ctx = ListCallerExecutionContext(caller, self.declared)
        return self.inner.extend(list(map(lambda x: self.checker.check_and_wrap(x, ctx), iterable)))

    def __iter__(self):
        return TypedListIterator(self)

    def __class_getitem__(self, item):
        return self.inner.__class_getitem__(item)

    def __reversed__(self, *args, **kwargs):
        return self.inner.__reversed__(*args, **kwargs)

class TypedListIterator:
    inner: TypedList
    index: int

    def __init__(self, inner):
        super().__init__()
        self.inner = inner
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self.inner):
            raise StopIteration

        ret = self.inner[self.index]
        self.index += 1
        return ret



import collections.abc
import inspect
from collections.abc import Iterator
from typing import Any, Optional
from typing import Iterator as OtherIterator

from untypy.error import UntypyTypeError, UntypyAttributeError, Location
from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext
from untypy.util import CompoundTypeExecutionContext

IteratorTypeA = type(Iterator[int])
IteratorTypeB = type(OtherIterator[int])


class IteratorFactory(TypeCheckerFactory):
    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if type(annotation) in [IteratorTypeA, IteratorTypeB] and annotation.__origin__ == collections.abc.Iterator:
            if len(annotation.__args__) != 1:
                raise ctx.wrap(UntypyAttributeError(f"Expected 1 type arguments for iterator."))

            inner = ctx.find_checker(annotation.__args__[0])
            if inner is None:
                raise ctx.wrap(UntypyAttributeError(f"The inner type of the iterator could not be resolved."))
            return IteratorChecker(inner)
        else:
            return None


class IteratorChecker(TypeChecker):
    inner: TypeChecker

    def __init__(self, inner: TypeChecker):
        self.inner = inner

    def may_be_wrapped(self) -> bool:
        return True

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if not hasattr(arg, '__next__') or not hasattr(arg, '__iter__'):
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

        me = self
        ctx = TypedIteratorExecutionContext(self.inner, arg, ctx)

        def wrapper():
            for item in arg:
                yield me.inner.check_and_wrap(item, ctx)

        return wrapper()

    def describe(self) -> str:
        return f"Iterator[{self.inner.describe()}]"

    def base_type(self) -> list[Any]:
        return [IteratorTypeA]


class TypedIteratorExecutionContext(CompoundTypeExecutionContext):
    iter: Iterator[Any]

    def __init__(self, inner: TypeChecker, iter: Iterator[Any], upper: ExecutionContext):
        self.iter = iter
        super().__init__(upper, [inner], 0)

    def name(self) -> str:
        return "Iterator"

    def responsable(self) -> Optional[Location]:
        try:
            if hasattr(self.iter, 'gi_frame'):
                return Location(
                    file=inspect.getfile(self.iter.gi_frame),
                    line_no=inspect.getsourcelines(self.iter.gi_frame)[1],
                    line_span=len(inspect.getsourcelines(self.iter.gi_frame)[0]),
                )
        except OSError:  # this call does not work all the time
            pass
        except TypeError:
            pass
        return None



import ast
from typing import *


class ReturnTraceManager:
    """
    Stores file & line_no to every return idx
    """
    def __init__(self):
        self.lst = []

    def next_id(self, reti : ast.Return, file: str) -> int:
        i = len(self.lst)
        self.lst.append((file, reti.lineno))
        return i

    def get(self, idx : int) -> (str, int):
        return self.lst[idx]

GlobalReturnTraceManager = ReturnTraceManager()
reti_loc: int = -1


def before_return(idx : int):
    global reti_loc
    reti_loc = idx


def get_last_return() -> (str, int):
    global reti_loc
    if reti_loc < 0:
        return ("<nothing>", 0) # this will never match any real location

    # Note: this location is only used if it is in the span of the located function.
    # See ReturnExecutionContext
    return GlobalReturnTraceManager.get(reti_loc)


class ReturnTracesTransformer(ast.NodeTransformer):

    def __init__(self, file: str, manager=GlobalReturnTraceManager):
        self.file = file
        self.manager = manager

    def generic_visit(self, node) -> Any:
        # See https://docs.python.org/3/library/ast.html
        stmt_types = ["body", "orelse", "finalbody"]

        for stmt_type in stmt_types:
            if not hasattr(node, stmt_type):
                continue
            statements : list = getattr(node, stmt_type)

            if not isinstance(statements, Iterable):
                # skip lambda expressions
                continue

            inserts = []
            for i,s in enumerate(statements):
                if type(s) is ast.Return:
                    inserts.append((i, s))

            # start at end so the early indexes still fit
            inserts.reverse()

            for index, reti in inserts:
                n = ast.Expr(value=ast.Call(
                    func=ast.Attribute(value=ast.Name(id='untypy', ctx=ast.Load()), attr='_before_return', ctx=ast.Load()),
                    args=[ast.Constant(value=self.manager.next_id(reti, self.file))],
                    keywords=[])
                )
                statements.insert(index, n)

        super(ast.NodeTransformer, self).generic_visit(node)




from __future__ import annotations

import inspect
from enum import Enum
import os
from typing import Any, Optional, Tuple, Iterable


def readFile(path):
    try:
        with open(path, encoding='utf-8') as f:
            return f.read()
    except UnicodeDecodeError:
        with open(path) as f:
            return f.read()

def relpath(p):
    # relpath might throw an exception, at least on windows
    try:
        return os.path.relpath(p)
    except:
        return p

class Location:
    file: str
    line_no: int
    line_span : int
    source_lines: Optional[str]

    def __init__(self, file: str, line_no: int, line_span : int):
        self.file = file
        self.line_no = line_no
        self.line_span = line_span
        self.source_lines = None

    def source(self) -> Optional[str]:
        if self.source_lines is None:
            try:
                self.source_lines = readFile(self.file)
            except OSError:
                self.source_lines = ''
        return self.source_lines

    def source_lines_span(self) -> Optional[str]:
        # This is still used for unit testing

        source = self.source()
        if source is None:
            return None

        buf = ""

        for i, line in enumerate(source.splitlines()):
            if (i + 1) in range(self.line_no, self.line_no + self.line_span):
                buf += f"\n{line}"

        return buf

    def __str__(self):
        return f"{relpath(self.file)}:{self.line_no}"

    def formatWithCode(self):
        buf = str(self)
        source = self.source()
        if not source:
            return buf
        lines = source.splitlines()
        idx = self.line_no - 1
        if idx < 0 or idx > len(lines):
            return buf
        else:
            return buf + '\n  | ' + lines[idx]

    def __repr__(self):
        return f"Location(file={self.file.__repr__()}, line_no={self.line_no.__repr__()}, line_span={self.line_span})"

    def __eq__(self, other):
        if not isinstance(other, Location):
            return False
        return self.file == other.file and self.line_no == other.line_no

    @staticmethod
    def from_code(obj) -> Location:
        try:
            return Location(
                file=inspect.getfile(obj),
                line_no=inspect.getsourcelines(obj)[1],
                line_span=len(inspect.getsourcelines(obj)[0]),
            )
        except Exception:
            return Location(
                file=inspect.getfile(obj),
                line_no=1,
                line_span=1,
            )

    @staticmethod
    def from_stack(stack) -> Location:
        if isinstance(stack, inspect.FrameInfo):
            try:
                return Location(
                    file=stack.filename,
                    line_no=stack.lineno,
                    line_span=1
                )
            except Exception:
                return Location(
                    file=stack.filename,
                    line_no=stack.lineno,
                    line_span=1
                )
        else:  # assume sys._getframe(...)
            try:
                return Location(
                    file=stack.f_code.co_filename,
                    line_no=stack.f_lineno,
                    line_span=1
                )
            except Exception:
                return Location(
                    file=stack.f_code.co_filename,
                    line_no=stack.f_lineno,
                    line_span=1
                )

    def __contains__(self, other: Location):
        file, line = (other.file, other.line_no)
        return self.file == file and line in range(self.line_no, self.line_no + self.line_span)

    def narrow_in_span(self, reti_loc: Tuple[str, int]):
        """
        Use new Location if inside of span of this Location
        :param reti_loc: filename and line_no
        :return: a new Location, else self
        """
        file, line = reti_loc
        if self.file == file and line in range(self.line_no, self.line_no + self.line_span):
            return Location(
                file=file,
                line_no=line,
                line_span=1
            )
        else:
            return self


class Frame:
    type_declared: str
    indicator_line: str

    declared: Optional[Location]
    responsable: Optional[Location]

    responsibility_type: Optional[ResponsibilityType]

    def __init__(self, type_declared: str, indicator_line: Optional[str],
                 declared: Optional[Location], responsable: Optional[Location]):

        self.type_declared = type_declared
        if indicator_line is None:
            indicator_line = '^' * len(type_declared)
        self.indicator_line = indicator_line
        self.declared = declared
        self.responsable = responsable

    def __str__(self):
        buf = f"in: {self.type_declared}\n" \
              f"    {self.indicator_line}\n"

        if self.responsable is not None:
            buf += f"{self.responsable.file}:{self.responsable.line_no}:\n" \
                   f"{self.responsable.source_line}\n" \
                   f"\n"
        return buf


class ResponsibilityType(Enum):
    IN = 0
    OUT = 1

    def invert(self):
        if self is ResponsibilityType.IN:
            return ResponsibilityType.OUT
        else:
            return ResponsibilityType.IN

def joinLines(l: Iterable[str]) -> str:
    return '\n'.join([x.rstrip() for x in l])

# Note: the visual studio code plugin uses the prefixes "caused by: " and "declared at: "
# for finding source locations. Do not change without changing the plugin code!!
CAUSED_BY_PREFIX = "caused by: "
DECLARED_AT_PREFIX = "declared at: "

def formatLocations(prefix: str, locs: list[Location]) -> str:
    return joinLines(map(lambda s: prefix + str(s), locs))

# All error types must be subclasses from UntypyError.
class UntypyError:
    def simpleName(self):
        raise Exception('abstract method')

NO_GIVEN = object()

class UntypyTypeError(TypeError, UntypyError):
    given: Any  # NO_GIVEN if not present
    header: str
    expected: Optional[str]
    frames: list[Frame]
    notes: list[str]
    previous_chain: Optional[UntypyTypeError]
    responsibility_type: ResponsibilityType

    def __init__(self,
                 given: Any = NO_GIVEN,
                 expected: str = None,
                 frames: list[Frame] = [],
                 notes: list[str] = [],
                 previous_chain: Optional[UntypyTypeError] = None,
                 responsibility_type: ResponsibilityType = ResponsibilityType.IN,
                 header: str = ''):
        self.responsibility_type = responsibility_type
        self.given = given
        self.expected = expected
        self.frames = frames.copy()
        for frame in self.frames:
            if frame.responsibility_type is None:
                frame.responsibility_type = responsibility_type
        self.notes = notes.copy()
        self.previous_chain = previous_chain
        self.header = header
        super().__init__('\n' + self.__str__())

    def simpleName(self):
        return 'TypeError'

    def next_type_and_indicator(self) -> Tuple[str, str]:
        if len(self.frames) >= 1:
            frame = self.frames[-1]
            return frame.type_declared, frame.indicator_line
        else:
            n = 0
            if self.expected:
                n = len(self.expected)
            return self.expected, "^" * n

    def with_frame(self, frame: Frame) -> UntypyTypeError:
        frame.responsibility_type = self.responsibility_type
        return UntypyTypeError(self.given, self.expected, self.frames + [frame],
                               self.notes, self.previous_chain, self.responsibility_type,
                               self.header)

    def with_previous_chain(self, previous_chain: UntypyTypeError):
        return UntypyTypeError(self.given, self.expected, self.frames,
                               self.notes, previous_chain, self.responsibility_type, self.header)

    def with_note(self, note: str):
        return UntypyTypeError(self.given, self.expected, self.frames,
                               self.notes + [note], self.previous_chain, self.responsibility_type,
                               self.header)

    def with_inverted_responsibility_type(self):
        return UntypyTypeError(self.given, self.expected, self.frames,
                               self.notes, self.previous_chain, self.responsibility_type.invert(),
                               self.header)

    def with_header(self, header: str):
        return UntypyTypeError(self.given, self.expected, self.frames,
                               self.notes, self.previous_chain, self.responsibility_type, header)

    def last_responsable(self):
        for f in reversed(self.frames):
            if f.responsable is not None and f.responsibility_type is ResponsibilityType.IN:
                return f.responsable
        return None

    def last_declared(self):
        for f in reversed(self.frames):
            if f.declared is not None:
                return f.declared
        return None

    def __str__(self):
        declared_locs = []
        responsable_locs = []

        for f in self.frames:
            if f.responsable is not None and f.responsibility_type is ResponsibilityType.IN:
                s = f.responsable.formatWithCode()
                if s not in responsable_locs:
                    responsable_locs.append(s)
            if f.declared is not None:
                s = str(f.declared)
                if s not in declared_locs:
                    declared_locs.append(str(f.declared))

        cause = formatLocations(CAUSED_BY_PREFIX, responsable_locs)
        declared = formatLocations(DECLARED_AT_PREFIX, declared_locs)

        (ty, ind) = self.next_type_and_indicator()

        notes = joinLines(self.notes)
        if notes:
            notes = notes + "\n"

        if self.previous_chain is None:
            previous_chain = ""
            preHeader = self.header or 'got value of wrong type'
            postHeader = ''
        else:
            previous_chain = self.previous_chain.__str__().strip()
            preHeader = ''
            postHeader = self.header
        if postHeader:
            postHeader = postHeader + "\n"
        if preHeader:
            preHeader = preHeader + "\n"
        if previous_chain:
            previous_chain = previous_chain + "\n\n"

        ctx = ""
        if self.expected != ty:
            ctx = f"context: {ty.rstrip()}"
            ind = ind.rstrip()
            if ind:
                ctx = f"{ctx}\n         {ind}"
        if ctx:
            ctx = ctx + "\n"
        given = None if self.given is NO_GIVEN else repr(self.given)
        expected = None if self.expected is None else self.expected.strip()
        if expected is not None and expected != 'None':
            expected = f'value of type {expected}'
        if given is not None:
            given = f"given:    {given.rstrip()}\n"
        else:
            given = ""
        if expected is not None:
            expected = f"expected: {expected}\n"
        else:
            expected = ""
        if notes and (given or expected):
            notes += "\n"
        return (f"""{preHeader}{previous_chain}{postHeader}{notes}{given}{expected}
{ctx}{declared}
{cause}""")


class UntypyAttributeError(AttributeError, UntypyError):

    def __init__(self, message: str, locations: list[Location] = []):
        self.message = message
        self.locations = locations.copy()
        super().__init__(self.__str__())

    def simpleName(self):
        return 'AttributeError'

    def with_location(self, loc: Location) -> UntypyAttributeError:
        return type(self)(self.message, self.locations + [loc]) # preserve type

    def __str__(self):
        return f"{self.message}\n{formatLocations(DECLARED_AT_PREFIX, self.locations)}"


class UntypyNameError(UntypyAttributeError, UntypyError):
    def simpleName(self):
        return 'NameError'



import ast
import inspect
import typing

from untypy.error import UntypyNameError, UntypyAttributeError
from untypy.interfaces import CreationContext, WrappedFunction
from untypy.util.source_utils import DisplayMatrix

RULES = []


def _default_resolver(item):
    return typing.get_type_hints(item, include_extras=True)


def get_type_hints(item, ctx: CreationContext, resolver=_default_resolver):
    try:
        # SEE: https://www.python.org/dev/peps/pep-0563/#id7
        return resolver(item)
    except NameError as ne:
        org = WrappedFunction.find_original(item)
        if inspect.isclass(org):
            raise ctx.wrap(UntypyNameError(
                f"{ne}.\nType annotation inside of class '{qualname(org)}' could not be resolved."
            ))
        else:
            raise ctx.wrap(UntypyNameError(
                f"{ne}.\nType annotation of function '{qualname(org)}' could not be resolved."
            ))
    except Exception as e:
        # Try to find better cause in analyse
        raise analyse(item, ctx, e)


def analyse(item, ctx: CreationContext, e) -> UntypyAttributeError:
    org = WrappedFunction.find_original(item)
    source = inspect.getsource(item)
    fn_ast = ast.parse(source)

    for node in map_str_to_ast(fn_ast.body[0].args, fn_ast.body[0].returns):
        for rule in RULES:
            rule_result = rule(node)
            if rule_result:
                # Got a Match
                (n, message) = rule_result
                display = DisplayMatrix(source)
                display.write((n.col_offset - 1, n.lineno),
                              " " + "^" * (n.end_col_offset - n.col_offset) + " - " + message)
                return ctx.wrap(
                    UntypyAttributeError(f"Type annotation of function '{qualname(org)}' could not be resolved:\n"
                                         f"{e}\n"
                                         f"\n{display}")
                )

    return ctx.wrap(
        UntypyAttributeError(f"Type annotation of function '{qualname(org)}' could not be resolved:\n"
                             f"{e}\n"))


# some type annotations may repr. as strings
def map_str_to_ast(*nodes):
    for outer_node in nodes:
        for node in ast.walk(outer_node):
            yield node
            if isinstance(node, ast.Constant) and isinstance(node.value, str):
                try:
                    for inode in ast.walk(ast.parse(node.value, mode='eval').body):
                        if hasattr(inode, 'lineno'):
                            inode.lineno += node.lineno - 1
                            inode.col_offset += node.col_offset + 1
                            inode.end_col_offset += node.col_offset + 1
                        yield inode
                except SyntaxError:
                    # may not a forward ref
                    pass


# Rules
# For Ast-Nodes see: https://docs.python.org/3/library/ast.html
def rule_wrong_parentheses(item):
    if isinstance(item, ast.Call) and _traverse(item, 'func.id') and _traverse(item, 'args.0'):
        inner = ", ".join(map(ast.unparse, _traverse(item, 'args')))
        return (item, f"Did you mean: '{_traverse(item, 'func.id')}[{inner}]'?")


RULES.append(rule_wrong_parentheses)


# Helpers
def _traverse(item, path):
    """
    Traverse item.
    Else return None.
    Path is a str separated by '.'
    """

    if isinstance(path, str):
        return _traverse(item, path.split("."))

    if len(path) == 0:
        return item

    head = path[0]
    tail = path[1:]

    if hasattr(item, head):
        return _traverse(getattr(item, head), tail)
    elif isinstance(item, list):
        try:
            idx = int(head)
            if len(item) > idx >= 0:
                return _traverse(item[idx], tail)
        except ValueError:
            return None
    else:
        return None


def qualname(typ):
    if hasattr(typ, '__qualname__'):
        return typ.__qualname__
    elif hasattr(typ, '__name__'):
        return typ.__name__
    else:
        return str(typ)



from collections.abc import Iterator, Iterable
from typing import TypeVar, Optional, Any, Generic, Dict, List, Set, Tuple, Protocol

from untypy.error import UntypyAttributeError, UntypyTypeError
from untypy.impl.protocol import ProtocolChecker
from untypy.impl.wrappedclass import WrappedType
from untypy.interfaces import TypeCheckerFactory, TypeChecker, CreationContext, ExecutionContext
from untypy.util import ReplaceTypeExecutionContext

A = TypeVar("A")
B = TypeVar("B")


class WDictLike(Protocol[A, B]):
    """
    This protocol implements a subset of dict.
    It exists solly to prevent an recursion issue
    inside of WDict
    """

    def __iter__(self) -> Iterator[A]:
        pass

    def __getitem__(self, item: A) -> B:
        pass


K = TypeVar("K")
V = TypeVar("V")


# See: https://docs.python.org/3/library/stdtypes.html#typesmapping
class WDict(Generic[K, V], dict):
    def clear(self) -> None:
        pass

    # Cannot Typecheck Copy -> Leads to endless recursion in "UntypyInterfaces"
    # def copy(self) -> dict[K,V]:
    #     pass

    def get(self, key: K, default: Optional[V] = None) -> Optional[V]:
        pass

    def items(self) -> Iterable[Tuple[K, V]]:
        pass

    def keys(self) -> Iterable[K]:
        pass

    def pop(self, k: K, default: Optional[V] = None) -> Optional[V]:
        pass

    def popitem(self) -> Tuple[K, V]:
        pass

    # Miss-match See: https://github.com/skogsbaer/write-your-python-program/issues/19
    def setdefault(self, key: K, default: V) -> V:
        pass

    def update(self, *E: Iterable[WDictLike[K, V]], **F: Optional[WDictLike[K, V]]) -> Any:
        pass

    def values(self) -> Iterable[V]:
        pass

    def __contains__(self, key: K) -> bool:
        pass

    def __delitem__(self, key: K) -> None:
        pass

    def __iter__(self) -> Iterator[K]:
        pass

    def __len__(self) -> int:
        pass

    # Untypy does not support generic functions :/
    # def __or__(self, other : dict[I, J]) -> dict[Union[K,I], Union[V,J]]:
    #     pass

    def __reversed__(self) -> Iterator[K]:
        pass

    # Untypy does not support generic functions :/
    # def __ror__(self, other : dict[I, J]) -> dict[Union[K,I], Union[V,J]]:
    #     """ Return value|self. """
    #     pass

    def __getitem__(self, item: K) -> V:
        pass

    def __setitem__(self, key: K, value: V) -> None:
        pass


I = TypeVar("I")


class WList(Generic[I], list):
    def __getitem__(self, item: int) -> I:
        pass

    def __setitem__(self, key: int, value: I) -> None:
        pass


I = TypeVar("I")


class WSet(Generic[I], set):

    def add(self, other: I) -> None:
        pass

    def clear(self) -> None:
        pass

    def discard(self, elem: I):
        pass

    def pop(self) -> I:
        pass

    def remove(self, elem: I) -> None:
        pass

    def update(self, *others: Tuple[Iterable[I], ...]) -> None:
        pass

    # This method returns `NotImplemented`, i don't know why.
    def __ior__(self, *others: Tuple[Iterable[I], ...]) -> Any:
        pass

    def __contains__(self, item: I) -> bool:
        pass

    def __iter__(self) -> Iterator[I]:
        pass

    def __len__(self) -> int:
        pass

    # Only removes elements. No checking needed. Argument type is set Any
    #
    # def intersection_update(self, others: Iterator[set[Any]]) -> None:
    #     pass
    #
    # def __iand__(self, other: set[Any]) -> None:
    #     pass
    #
    # def difference_update(self, others: Iterator[set[Any]]) -> None:
    #     pass
    #
    # def __isub__(self, other: set[Any]) -> None:
    #     pass

    # Mutable meth
    # symmetric_difference_update
    # __ixor__

    # Recursion
    # copy

    # Immutable generic meth
    # difference
    # intersection
    # isdisjoint
    # issubset
    # issuperset
    # symmetric_difference(self, other)
    # union(self, other)
    # __and__
    # __or__
    # __rand__
    # __reduce__ ???
    # __ror__
    # __rsub__
    # __rxor__
    # __sub__
    # __xor__


I = TypeVar("I")


class WIterable(Generic[I]):
    def __iter__(self) -> Iterator[I]:
        pass


InterfaceMapping = {
    dict: (WDict,),
    Dict: (WDict,),
    WDictLike: (WDictLike[K, V],),
    list: (WList,),
    List: (WList,),
    set: (WSet,),
    Set: (WSet,),
    Iterable: (WIterable,),
}

InterfaceFactoryCache = {}


class InterfaceFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if hasattr(annotation, '__origin__') and hasattr(annotation,
                                                         '__args__') and annotation.__origin__ in InterfaceMapping:
            (protocol,) = InterfaceMapping[annotation.__origin__]
            bindings = protocol.__parameters__  # args of Generic super class
            origin = annotation.__origin__

            inner_checkers = []
            for param in annotation.__args__:
                ch = ctx.find_checker(param)
                if ch is None:
                    raise UntypyAttributeError(f"Could not resolve annotation {param} inside of {annotation}")
                inner_checkers.append(ch)
            if len(inner_checkers) != len(bindings):
                raise UntypyAttributeError(f"Expected {len(bindings)} type arguments inside of {annotation}")

            name = f"{origin.__name__}[" + (', '.join(map(lambda t: t.describe(), inner_checkers))) + "]"

            bindings = dict(zip(bindings, annotation.__args__))
            ctx = ctx.with_typevars(bindings)
            if type(origin) == type:
                template = WrappedType(protocol, ctx.with_typevars(bindings), name=name, implementation_template=origin,
                                       declared=ctx.declared_location())
                return InterfaceChecker(origin, template, name)
            else:
                # type(origin) == collection.abc.ABCMeta
                return ProtocolChecker(protocol, ctx, altname=name)
        else:
            return None


class InterfaceChecker(TypeChecker):

    def __init__(self, origin, template, name):
        self.origin = origin
        self.template = template
        self.name = name
        pass

    def may_change_identity(self) -> bool:
        return True

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if not issubclass(type(arg), self.origin):
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

        if hasattr(arg, '_WrappedClassFunction__inner'):
            # Prevent Double Wrapping
            arg = arg._WrappedClassFunction__inner

        instance = self.template.__new__(self.template)
        instance._WrappedClassFunction__inner = arg
        instance._WrappedClassFunction__return_ctx = ReplaceTypeExecutionContext(ctx, self.name)
        return instance

    def describe(self) -> str:
        return self.name



import inspect
import sys
from types import ModuleType
from typing import Any, Callable, Union, Optional

from untypy.util.display import format_argument_values
from untypy.error import Location, UntypyAttributeError, UntypyTypeError
from untypy.impl.any import SelfChecker, AnyChecker
from untypy.interfaces import TypeChecker, CreationContext, ExecutionContext, WrappedFunction, \
    WrappedFunctionContextProvider
from untypy.util import ArgumentExecutionContext, ReturnExecutionContext


def find_signature(member, ctx: CreationContext):
    signature = inspect.signature(member)

    checkers = {}
    for key in signature.parameters:
        if key == 'self':
            checkers[key] = SelfChecker()
        else:
            param = signature.parameters[key]
            if param.annotation is inspect.Parameter.empty:
                checkers[key] = AnyChecker()
                continue

            checker = ctx.find_checker(param.annotation)
            if checker is None:
                checkers[key] = AnyChecker()
                continue
            checkers[key] = checker

    if signature.return_annotation is inspect.Parameter.empty:
        checkers['return'] = AnyChecker()
    else:
        return_checker = ctx.find_checker(signature.return_annotation)
        if return_checker is None:
            raise ctx.wrap(UntypyAttributeError(f"\n\tUnsupported Type Annotation: {signature.return_annotation}\n"
                                                f"for Return Value of function {member.__name__}\n"))
        checkers['return'] = return_checker
    return signature, checkers


def wrap_arguments(signature: inspect.signature, checker: dict[str, TypeChecker],
                   ctxprv: WrappedFunctionContextProvider, args, kwargs):
    bindings = signature.bind(*args, **kwargs)
    bindings.apply_defaults()
    for name in bindings.arguments:
        check = checker[name]
        ctx = ctxprv(name)
        bindings.arguments[name] = check.check_and_wrap(bindings.arguments[name], ctx)
    return bindings.args, bindings.kwargs


def WrappedType(template: Union[type, ModuleType], ctx: CreationContext, *,
                implementation_template: Union[type, ModuleType, None] = None,
                create_type: Optional[type] = None,
                name: Optional[str] = None,
                declared: Optional[Location] = None):
    blacklist = ['__class__', '__delattr__', '__dict__', '__dir__',
                 '__doc__', '__getattribute__', '__getattr__', '__init_subclass__',
                 '__new__', '__setattr__', '__subclasshook__', '__weakref__']

    create_fn = None
    if create_type is not None:
        create_fn = lambda: create_type.__new__(create_type)

    if type(template) is type and create_type is None:
        create_fn = lambda: template.__new__(template)

    if implementation_template is None:
        implementation_template = template

    if create_fn is None:
        def raise_err():
            raise TypeError("This is not a Callable")

        create_fn = raise_err

    list_of_attr = dict()
    for attr in dir(template):
        if attr in blacklist:
            continue

        original = getattr(template, attr)
        if type(original) == type:  # Note: Order matters, types are also callable
            if type(template) is type:
                list_of_attr[attr] = WrappedType(original, ctx)

        elif callable(original):
            try:
                (signature, checker) = find_signature(original, ctx)
                implementation_fn = getattr(implementation_template, attr)
                if implementation_fn is not None:
                    list_of_attr[attr] = WrappedClassFunction(implementation_fn, signature, checker,
                                                              create_fn=create_fn, declared=declared).build()
            except ValueError as e:
                # this fails sometimes on built-ins.
                # "ValueError: no signature found for builtin"
                pass
    out = None
    if type(template) is type:
        if name is None:
            name = f"{template.__name__}Wrapped"
        out = type(name, (template,), list_of_attr)
    elif inspect.ismodule(template):
        out = type("WrappedModule", (), list_of_attr)

    return out


class WrappedClassFunction(WrappedFunction):
    def __init__(self,
                 inner: Callable,
                 signature: inspect.Signature,
                 checker: dict[str, TypeChecker], *,
                 create_fn: Callable[[], Any],
                 declared: Optional[Location] = None):
        self.inner = inner
        self.signature = signature
        self.checker = checker
        self.create_fn = create_fn
        self._declared = declared
        self.fc = None
        if hasattr(self.inner, "__fc"):
            self.fc = getattr(self.inner, "__fc")

    def build(self):
        fn = self.inner
        name = fn.__name__

        def wrapper_cls(*args, **kwargs):
            caller = sys._getframe(1)
            (args, kwargs, bindings) = self.wrap_arguments(
                lambda n: ArgumentExecutionContext(wrapper_cls, caller, n, declared=self.declared()),
                args, kwargs)
            ret = fn(*args, **kwargs)
            return self.wrap_return(ret, bindings, ReturnExecutionContext(self))

        def wrapper_self(me, *args, **kwargs):
            if name == '__init__':
                me.__return_ctx = None
                me.__inner = self.create_fn()
            caller = sys._getframe(1)
            (args, kwargs, bindings) = self.wrap_arguments(
                lambda n: ArgumentExecutionContext(wrapper_self, caller, n, declared=self.declared()),
                (me.__inner, *args), kwargs)
            ret = fn(*args, **kwargs)
            if me.__return_ctx is None:
                return self.wrap_return(ret, bindings, ReturnExecutionContext(self))
            else:
                return self.wrap_return(ret, bindings, me.__return_ctx)

        if inspect.iscoroutine(self.inner):
            raise UntypyAttributeError("Async Functions are currently not supported.")
        else:
            if 'self' in self.checker:
                w = wrapper_self
            else:
                w = wrapper_cls

        setattr(w, '__wrapped__', fn)
        setattr(w, '__name__', fn.__name__)
        setattr(w, '__signature__', self.signature)
        setattr(w, '__wf', self)
        return w

    def get_original(self):
        return self.inner

    def wrap_arguments(self, ctxprv: WrappedFunctionContextProvider, args, kwargs):
        try:
            bindings = self.signature.bind(*args, **kwargs)
        except TypeError as e:
            err = UntypyTypeError(header=str(e))
            raise ctxprv("").wrap(err)

        bindings.apply_defaults()
        if self.fc is not None:
            self.fc.prehook(bindings, ctxprv)
        for name in bindings.arguments:
            check = self.checker[name]
            ctx = ctxprv(name)
            bindings.arguments[name] = check.check_and_wrap(bindings.arguments[name], ctx)
        return bindings.args, bindings.kwargs, bindings

    def wrap_return(self, ret, bindings, ctx: ExecutionContext):
        check = self.checker['return']
        if self.fc is not None:
            self.fc.posthook(ret, bindings, ctx)
        return check.check_and_wrap(ret, ctx)

    def describe(self) -> str:
        fn = WrappedFunction.find_original(self.inner)
        return f"{fn.__name__}" + str(self.signature)

    def checker_for(self, name: str) -> TypeChecker:
        return self.checker[name]

    def declared(self) -> Location:
        if self._declared is None:
            return WrappedFunction.find_location(self.inner)
        else:
            return self._declared



# FIXME: make exceptions nicer
import untypy
import typing
import dataclasses
import inspect
import types

_DEBUG = False
def _debug(s):
    if _DEBUG:
        print('[DEBUG] ' + s)

# Types
Any = typing.Any
Optional = typing.Optional
Union = typing.Union
Iterable = typing.Iterable
Iterator = typing.Iterator
Sequence = typing.Sequence
Generator = typing.Generator
ForwardRef = typing.ForwardRef

Mapping = typing.Mapping

Callable = typing.Callable

dataclass = dataclasses.dataclass

unchecked = untypy.unchecked

intPositive = typing.Annotated[int, lambda i: i > 0, 'intPositive']
nat = typing.Annotated[int, lambda i: i >= 0, 'nat']
intNonNegative = typing.Annotated[int, lambda i: i >= 0, 'intNonNegative']
intNonPositive = typing.Annotated[int, lambda i: i <= 0, 'intNonPositive']
intNegative = typing.Annotated[int, lambda i: i < 0, 'intNegative']

floatPositive = typing.Annotated[float, lambda x: x > 0, 'floatPositive']
floatNonNegative = typing.Annotated[float, lambda x: x >= 0, 'floatNonNegative']
floatNegative = typing.Annotated[float, lambda x: x < 0, 'floatNegative']
floatNonPositive = typing.Annotated[float, lambda x: x <= 0, 'floatNonPositive']

T = typing.TypeVar('T')
U = typing.TypeVar('U')
V = typing.TypeVar('V')


class Literal(types.GenericAlias):
    """
    Behaves like typing.Literal, but allows isinstance checks.
    """
    # Note: Cannot subclass Literal
    # It is also not possible to overwrite '__instancecheck__'
    # We musst inherit from GenericAlias, or else this can not be used
    # as part of other types.

    def __class_getitem__(cls, items):
        # Multiple __class_getitem__ are given as a tuple,
        # but if there is only a single argument, it is not wrapped.
        if not isinstance(items, tuple):
            items = (items,)
        # flattening
        # cannot reuse typing.Literal for flattening,
        # as it does not recognize this type
        args_set = []
        for i in items:
            if hasattr(i, '__origin__') and hasattr(i, '__args__') and i.__origin__ in [typing.Literal]:
                for arg in i.__args__:
                    if arg not in args_set:
                        args_set.append(arg)
            elif i not in args_set:
                args_set.append(i)

        return Literal(cls, tuple(args_set))

    def __instancecheck__(self, value):
        for arg in self.__args__:
            # typing.Literal checks for type(value) == type(arg),
            # however this does not work well with untypy.
            if isinstance(value, type(arg)) and arg == value:
                return True
        return False

    def __getattribute__(self, item):
        # types.GenericAlias does some undocumented subclassing magic,
        # we cannot set '__origin__', so we use __getattribute__
        # __origin__ should be typing.Literal, so it is recognize by untypy
        if item == '__origin__':
            return typing.Literal
        else:
            return types.GenericAlias.__getattribute__(self, item)

    def __eq__(self, other):
        if hasattr(other, '__origin__') and hasattr(other, '__args__') and other.__origin__ in [typing.Literal]:
            return set(self.__args__) == set(other.__args__)
        else:
            return False

    def __hash__(self):
        return hash(frozenset(self.__args__))

def _collectDataClassAttributes(cls):
    result = dict()
    for c in cls.mro():
        if hasattr(c, '__kind') and c.__kind == 'record' and hasattr(c, '__annotations__'):
            result = c.__annotations__ | result
    return result


def _patchDataClass(cls, mutable):
    fieldNames = [f.name for f in dataclasses.fields(cls)]
    setattr(cls, EQ_ATTRS_ATTR, fieldNames)

    if hasattr(cls, '__annotations__'):
        # add annotions for type checked constructor.
        cls.__kind = 'record'
        cls.__init__.__annotations__ = _collectDataClassAttributes(cls)
        cls.__init__.__original = cls # mark class as source of annotation
        cls.__init__ = untypy.typechecked(cls.__init__)

    if mutable:
        # prevent new fields being added
        fields = set(fieldNames)

        checker = {}
        # Note: Partial annotations are disallowed by untypy.typechecked(cls.__init__)
        #       So no handling in this code is required.
        for name in fields:
            if name in cls.__annotations__:
                # This the type is wrapped in an lambda expression to allow for Forward Ref.
                # Would the lambda expression be called at this moment, it may cause an name error
                # untypy.checker fetches the annotation lazily.
                checker[name] = untypy.checker(\
                    lambda cls=cls,name=name: typing.get_type_hints(cls, include_extras=True)[name],
                    cls)

        oldSetattr = cls.__setattr__
        def _setattr(obj, k, v):
            # Note __setattr__ also gets called in the constructor.
            if k in checker:
                oldSetattr(obj, k, checker[k](v))
            elif k in fields:
                oldSetattr(obj, k, v)
            else:
                raise AttributeError(f'Unknown attribute {k} for record {cls.__name__}')
        setattr(cls, "__setattr__", _setattr)
    return cls

def record(cls=None, mutable=False):
    def wrap(cls):
        newCls = dataclasses.dataclass(cls, frozen=not mutable)
        return _patchDataClass(newCls, mutable)
    # See if we're being called as @record or @record().
    if cls is None:
        # We're called with parens.
        return wrap
    else:
        # We're called as @dataclass without parens.
        return wrap(cls)

# Tests

_die = False

def setDieOnCheckFailures(b):
    _die = b

def _dieOnCheckFailures():
    return _die

_testCount = {'total': 0, 'failing': 0}
_checksEnabled = True

def initModule(enableChecks=True, quiet=False):
    global _checksEnabled
    _checksEnabled = enableChecks
    resetTestCount()

def resetTestCount():
    global _testCount
    _testCount = {'total': 0, 'failing': 0}

def printTestResults(prefix=''):
    total = _testCount['total']
    failing = _testCount['failing']
    if total == 0:
        pass
    elif failing == 0:
        print(f'{prefix}{total} Tests, alle erfolgreich :-)')
    else:
        print(f'{prefix}{total} Tests, {failing} Fehler :-(')
    return {'total': total, 'failing': failing}

def checkEq(actual, expected):
    return check(actual, expected, structuralObjEq=False)

def check(actual, expected, structuralObjEq=True, floatEqWithDelta=True):
    if not _checksEnabled:
        return
    global _testCount
    flags = {'structuralObjEq': structuralObjEq, 'floatEqWithDelta': floatEqWithDelta}
    matches = deepEq(actual, expected, **flags)
    _testCount = {
        'total': _testCount['total'] + 1,
        'failing': _testCount['failing'] + (0 if matches else 1)
    }
    if not matches:
        stack = inspect.stack()
        caller = stack[1] if len(stack) > 1 else None
        def fmt(x):
            if type(x) == str:
                return repr(x)
            else:
                return str(x)
        msg = f"{caller.filename}:{caller.lineno}: Erwartet wird {fmt(expected)}, aber das " \
            f"Ergebnis ist {fmt(actual)}"
        if _dieOnCheckFailures():
            raise Exception(msg)
        else:
            print("FEHLER in " + msg)

def uncoveredCase():
    stack = inspect.stack()
    caller = stack[1] if len(stack) > 1 else None
    raise Exception(f"{caller.filename}, Zeile {caller.lineno}: ein Fall ist nicht abgedeckt")

#
# Deep equality
#
def _isNumber(x):
    t = type(x)
    return (t is int or t is float)

def _seqEq(seq1, seq2, flags):
    if len(seq1) != len(seq2):
        return False
    for i, x in enumerate(seq1):
        y = seq2[i]
        if not deepEq(x, y, **flags):
            return False
    return True

def _dictEq(d1, d2, flags):
    ks1 = sorted(d1.keys())
    ks2 = sorted(d2.keys())
    if ks1 != ks2: # keys should be exactly equal
        return False
    for k in ks1:
        if not deepEq(d1[k], d2[k], **flags):
            return False
    return True

def _objToDict(o):
    d = {}
    attrs = dir(o)
    useAll = False
    if hasattr(o, EQ_ATTRS_ATTR):
        useAll = True
        attrs = getattr(o, EQ_ATTRS_ATTR)
    for n in attrs:
        if not useAll and n and n[0] == '_':
            continue
        x = getattr(o, n)
        d[n] = x
    #print(d)
    return d

def _objEq(o1, o2, flags):
    return _dictEq(_objToDict(o1), _objToDict(o2), flags)

_EPSILON = 0.00001
EQ_ATTRS_ATTR = '__eqAttrs__'

def _useFloatEqWithDelta(flags):
    return flags.get('floatEqWithDelta', False)

def _useStructuralObjEq(flags):
    return flags.get('structuralObjEq', False)

# Supported flags:
# - structuralObjEq: causes objects to be compared as dictionaries. The keys are by default
#   taken from dir(obj), if obj as the attribute __eqAttrs__, then the names listed there
#   are taken. Defaults to False.
# - floatEqWithDelta: compares floats by checking whether the difference is smaller than a
#   small delta. Setting this to True loses transitivity of eq.
def deepEq(v1, v2, **flags):
    """
    Computes deep equality of v1 and v2. With structuralObjEq=False, objects are compared
    by __eq__. Otherwise, objects are compared attribute-wise, only those attributes
    returned by dir that do not start with an underscore are compared.
    """
    # print(f'deepEq: v1={v1}, v2={v2}, flags={flags}')
    if v1 == v2:
        return True
    if _isNumber(v1) and _isNumber(v2) and _useFloatEqWithDelta(flags):
        diff = v1 - v2
        if abs(diff) < _EPSILON:
            return True
        else:
            return False
    if isinstance(v1, list):
        return isinstance(v2, list) and _seqEq(v1, v2, flags)
    if isinstance(v1, tuple):
        return isinstance(v2, tuple) and _seqEq(v1, v2, flags)
    if isinstance(v1, dict):
        return isinstance(v2, dict) and _dictEq(v1, v2, flags)
    if type(v1) == str:
        return False
    if hasattr(v1, '__class__'):
        if _useStructuralObjEq(flags):
            return _objEq(v1, v2, flags)
        else:
            return False # v1 == v2 already checked
    return False

# Additional functions and aliases

import math as moduleMath

math = moduleMath





import time
import threading
import writeYourProgram as _w
# Do not import tkinter at the top-level. Someone with no installation of tkinter should
# be able to user WYPP without drawing support.

@_w.record
class Size:
    width: float
    height: float

@_w.record
class Point:
    x: float
    y: float

ShapeKind = _w.Literal['ellipsis', 'rectangle']

Color = _w.Literal['red', 'green', 'blue', 'yellow', 'black', 'white']

@_w.record
class FixedShape:
    shape: ShapeKind
    color: Color
    size: Size
    position: Point

# coordinate system: origin is in the upper left corner
def _transformPoint(p, windowSize: Size):
    return (p[0] * 100 + windowSize.width/2, p[1] * -100 + windowSize.height/2)

def _renderFixedShape(prim: FixedShape, windowSize: Size, canvas):
    sz = prim.size
    pos = prim.position
    coords = [
        (pos.x - sz.width/2, pos.y - sz.height/2),
        (pos.x + sz.width/2, pos.y - sz.height/2),
        (pos.x + sz.width/2, pos.y + sz.height/2),
        (pos.x - sz.width/2, pos.y + sz.height/2)
    ]
    coords = [_transformPoint(p, windowSize) for p in coords]
    color = prim.color
    if prim.shape == 'rectangle':
        canvas.create_polygon(coords, fill=color, outline=color)
    elif prim.shape == 'ellipsis':
        upperLeft = coords[0]
        lowerRight = coords[2]
        canvas.create_oval(
            upperLeft[0], upperLeft[1],
            lowerRight[0], lowerRight[1],
            fill=color,
            outline=color
        )

def _drawCoordinateSystem(canvas, windowSize: Size):
    x = windowSize.width / 2
    y = windowSize.height / 2
    canvas.create_line(x, 0, x, windowSize.height, dash=(4,2))
    canvas.create_line(0, y, windowSize.width, y, dash=(4,2))

def drawFixedShapes(shapes: _w.Sequence[FixedShape],
                    withCoordinateSystem=False,
                    stopAfter=None) -> None:
    try:
        from tkinter import Tk, Frame, Canvas, BOTH
    except ImportError:
        import sys
        sys.stderr.write("Could not import tkinter. Please install the tkinter library.")
        raise
    class _WyppFrame(Frame):
        def __init__(self, diags, windowSize, withCoordinateSystem):
            super().__init__()
            self.master.title("Write Your Python Program")
            self.pack(fill=BOTH, expand=1)
            canvas = Canvas(self)
            if withCoordinateSystem:
                _drawCoordinateSystem(canvas, windowSize)
            for d in diags:
                _renderFixedShape(d, windowSize, canvas)
            canvas.pack(fill=BOTH, expand=1)
    root = Tk()
    winWidth = 800
    winHeight = 800
    ex = _WyppFrame(shapes, Size(winWidth, winHeight), withCoordinateSystem)
    root.geometry(f"{winWidth}x{winHeight}+50+50")
    root.lift()
    # bring window to front
    root.attributes('-topmost',True)
    root.after_idle(root.attributes,'-topmost',False)
    if stopAfter:
        def close():
            time.sleep(stopAfter)
            print('Closing drawing window automatically ...')
            root.destroy()
        t = threading.Thread(target=close, daemon=True)
        t.start()
    root.mainloop()



import inspect
from collections import namedtuple
from types import FunctionType
from typing import Callable, Protocol

from untypy.error import Location
from untypy.impl import DefaultCreationContext
from untypy.impl.bound_generic import WrappedGenericAlias
from untypy.impl.wrappedclass import WrappedType
from untypy.interfaces import WrappedFunction
from untypy.util.typedfunction import TypedFunctionBuilder

Config = namedtuple('PatchConfig', ['verbose', 'checkedprefixes'])
DefaultConfig = Config(verbose=False, checkedprefixes=[""])
not_patching = ['__class__']

GlobalPatchedList = set()


def patch_class(clas: type, cfg: Config):
    if clas in GlobalPatchedList:
        return clas
    GlobalPatchedList.add(clas)

    try:
        ctx = DefaultCreationContext(
            typevars=dict(),
            declared_location=Location(
                file=inspect.getfile(clas),
                line_no=inspect.getsourcelines(clas)[1],
                line_span=len(inspect.getsourcelines(clas)[0]),
            ), checkedpkgprefixes=cfg.checkedprefixes)
    except (TypeError, OSError) as e:  # Built in types
        ctx = DefaultCreationContext(
            typevars=dict(),
            declared_location=Location(
                file="<not found>",
                line_no=0,
                line_span=1
            ), checkedpkgprefixes=cfg.checkedprefixes,
        )

    setattr(clas, '__patched', True)

    is_protocol = hasattr(clas, 'mro') and Protocol in clas.mro()

    if hasattr(clas, '__class_getitem__') and not is_protocol:
        original = clas.__class_getitem__
        setattr(clas, '__class_getitem__', lambda *args: WrappedGenericAlias(original(*args), ctx))


def wrap_function(fn: FunctionType, cfg: Config) -> Callable:
    if len(inspect.getfullargspec(fn).annotations) > 0:
        if cfg.verbose:
            print(f"Patching Function: {fn.__name__}")
        return TypedFunctionBuilder(fn, DefaultCreationContext(
            typevars=dict(),
            declared_location=WrappedFunction.find_location(fn),
            checkedpkgprefixes=cfg.checkedprefixes, eval_context=fn.__globals__)).build()
    else:
        return fn


def wrap_class(a: type, cfg: Config) -> Callable:
    return WrappedType(a, DefaultCreationContext(
        typevars=dict(),
        declared_location=Location.from_code(a),
        checkedpkgprefixes=cfg.checkedprefixes))



from typing import Any, Optional

from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext


class AnyFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if annotation is Any:
            return AnyChecker()
        else:
            return None


class AnyChecker(TypeChecker):
    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        return arg

    def describe(self) -> str:
        return "Any"

    def base_type(self) -> type:
        return [Any]


class SelfChecker(TypeChecker):
    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        return arg

    def describe(self) -> str:
        return "Self"

    def base_type(self) -> type:
        return [Any]
